"""Zebra printer firmware"""

import logging
import os.path
import zipfile

logger = logging.getLogger(__name__)


class BadFirmwareError(Exception):
    """Invalid Zebra firmware image"""
    pass


class ZebraFirmware(object):
    """Zebra printer firmware"""
    # pylint: disable=too-few-public-methods

    def __init__(self, file):
        self.file = file
        with zipfile.ZipFile(file) as archive:
            zpls = [x for x in archive.namelist()
                    if os.path.splitext(x)[1].lower() == '.zpl']
            if not zpls:
                raise BadFirmwareError('No .zpl files in firmware')
            for zpl in zpls:
                logger.debug('found %s', zpl)
            if len(zpls) > 1:
                raise BadFirmwareError('Multiple .zpl files in firmware')
            self.zpl = zpls[0]
            with archive.open(self.zpl) as f:
                self.data = f.read()
            logger.debug('version %s', self.version)

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.file)

    def __bytes__(self):
        return self.data

    @property
    def version(self):
        """Firmware version"""
        return os.path.splitext(self.zpl)[0]



"""Zebra printer device connection"""

from abc import ABC, abstractmethod
import errno
import selectors
import socket
from urllib.parse import urlparse
import usb


class ZebraConnection(ABC):
    """A Zebra device connection"""

    def __init__(self, path=None):
        self.path = path

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.path)

    @abstractmethod
    def open(self):
        """Open connection"""
        pass

    @abstractmethod
    def close(self):
        """Close connection"""
        pass

    @abstractmethod
    def write(self, data):
        """Write data to connection"""
        pass

    @abstractmethod
    def read(self, size, timeout=None):
        """Read data from connection"""
        pass


class ZebraFileHandleConnection(ZebraConnection):
    """A Zebra file-handle device connection"""
    # pylint: disable=abstract-method

    def __init__(self, path=None):
        super().__init__(path)
        self.fh = None
        self.sel = selectors.DefaultSelector()

    def register(self, fh):
        """Register file handle"""
        self.fh = fh
        self.sel.register(self.fh, selectors.EVENT_READ)

    def close(self):
        self.sel.unregister(self.fh)
        self.fh.close()
        self.fh = None

    def write(self, data):
        return self.fh.write(data)

    def read(self, size, timeout=None):
        if not self.sel.select(timeout):
            return None
        return self.fh.read(size)


class ZebraFileConnection(ZebraFileHandleConnection):
    """A Zebra file device connection"""

    def open(self):
        super().open()
        self.register(open(self.path, 'r+b', buffering=0))


class ZebraNetworkConnection(ZebraFileHandleConnection):
    """A Zebra network device connection"""

    DEFAULT_PORT = 9100
    """Default port number for Zebra device"""

    def open(self):
        super().open()
        url = urlparse('//%s' % self.path)
        address = (url.hostname, url.port or self.DEFAULT_PORT)
        sock = socket.create_connection(address)
        self.register(sock.makefile('rwb', buffering=0))
        sock.close()


class ZebraUsbConnection(ZebraConnection):
    """A Zebra USB device connection"""

    DEFAULT_VENDOR = 0x0a5f
    """Default vendor ID for Zebra device"""

    def __init__(self, path=None):
        super().__init__(path)
        self.dev = None
        self.intf = None
        self.ep_in = None
        self.ep_out = None

    def open(self):
        intfs = lambda dev: (usb.util.find_descriptor(
            cfg, bInterfaceClass=usb.CLASS_PRINTER
        ) for cfg in dev)
        self.dev = usb.core.find(idVendor=self.DEFAULT_VENDOR,
                                 custom_match=lambda dev: any(intfs(dev)))
        if self.dev is None:
            raise FileNotFoundError("No Zebra printers found")
        self.intf = next(intfs(self.dev))
        ep_dir = lambda ep: usb.util.endpoint_direction(ep.bEndpointAddress)
        self.ep_in = next(x for x in self.intf
                          if ep_dir(x) == usb.ENDPOINT_IN)
        self.ep_out = next(x for x in self.intf
                           if ep_dir(x) == usb.ENDPOINT_OUT)
        try:
            self.dev.detach_kernel_driver(self.intf.bInterfaceNumber)
        except (NotImplementedError, usb.USBError):
            pass
        usb.util.claim_interface(self.dev, self.intf)
        self.dev.reset()

    def close(self):
        usb.util.release_interface(self.dev, self.intf)
        try:
            self.dev.attach_kernel_driver(self.intf.bInterfaceNumber)
        except (NotImplementedError, usb.USBError):
            pass
        usb.util.dispose_resources(self.dev)

    def write(self, data):
        return self.ep_out.write(data)

    def read(self, size, timeout=None):
        timeout_ms = None if timeout is None else int(timeout * 1000)
        try:
            raw = self.ep_in.read(size, timeout_ms)
        except usb.USBError as e:
            if e.errno == errno.ETIMEDOUT:
                return b''
            raise
        return bytes(raw)



"""Zebra configuration command line interface"""

import argparse
from collections import namedtuple
import logging
import os.path
import re
import colorlog
from .device import ZebraDevice
from .firmware import ZebraFirmware

LOG_FMT = '%(log_color)s%(levelname)s:%(name)s:%(message)s'
LOG_COLS = {**colorlog.default_log_colors, 'DEBUG': 'cyan'}


class NameValuePair(namedtuple('NameValuePair', ('name', 'value'))):
    """A name=value pair argument"""

    @classmethod
    def parse(cls, text):
        """Parse name=value pair from a string"""
        m = re.fullmatch(r'([\w\.]+)=(.+)', text)
        if not m:
            raise ValueError(text)
        return cls(*m.groups())


class ZebraCommand(object):
    """Configure Zebra printer"""

    loglevels = [logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG]

    def __init__(self, argv=None):
        self.args = self.parser().parse_args(argv)
        self.verbosity = (self.loglevels.index(logging.INFO) +
                          self.args.verbose - self.args.quiet)
        handler = colorlog.StreamHandler()
        handler.setFormatter(colorlog.ColoredFormatter(LOG_FMT,
                                                       log_colors=LOG_COLS))
        logging.basicConfig(level=self.loglevel, handlers=[handler])
        self.device = ZebraDevice(self.args.device, timeout=self.args.timeout)

    @classmethod
    def parser(cls, **kwargs):
        """Construct argument parser"""
        # pylint: disable=too-many-locals, unused-variable
        parser = argparse.ArgumentParser(description=cls.__doc__, **kwargs)
        common = argparse.ArgumentParser(add_help=False)
        common.add_argument('--verbose', '-v', action='count', default=0)
        common.add_argument('--quiet', '-q', action='count', default=0)
        common.add_argument('--device', '-d')
        common.add_argument('--timeout', '-t', type=float,
                            default=ZebraDevice.DEFAULT_TIMEOUT)
        cmds = parser.add_subparsers(title="subcommands", dest='subcommand')
        cmds.required = True

        getvar = cmds.add_parser('get', parents=[common])
        getvar.add_argument('variables', metavar='name', nargs='+')

        setvar = cmds.add_parser('set', parents=[common])
        setvar.add_argument('variables', metavar='name=value', nargs='+',
                            type=NameValuePair.parse)

        reset = cmds.add_parser('reset', parents=[common])

        restore = cmds.add_parser('restore', parents=[common])
        restore.add_argument('category')

        ls = cmds.add_parser('ls', parents=[common])

        rm = cmds.add_parser('rm', parents=[common])
        rm.add_argument('filename')

        mv = cmds.add_parser('mv', parents=[common])
        mv.add_argument('oldname')
        mv.add_argument('newname')

        cat = cmds.add_parser('cat', parents=[common])
        cat.add_argument('filename')

        pull = cmds.add_parser('pull', parents=[common])
        pull.add_argument('filename')
        pull.add_argument('--output', '-o')

        push = cmds.add_parser('push', parents=[common])
        push.add_argument('filename')
        push.add_argument('--input', '-i')

        upgrade = cmds.add_parser('upgrade', parents=[common])
        upgrade.add_argument('firmware')
        upgrade.add_argument('--check', action='store_true')

        wifi = cmds.add_parser('wifi', parents=[common])
        wifi.add_argument('essid')
        wifi.add_argument('--password', '-p')
        auth = wifi.add_mutually_exclusive_group()
        auth.add_argument('--wpa-psk', dest='auth', action='store_const',
                          const='wpa_psk')
        wifi.add_argument('--country')
        wifi.add_argument('--no-reset', dest='reset', action='store_false',
                          default=True)

        return parser

    @property
    def loglevel(self):
        """Log level"""
        return (self.loglevels[self.verbosity]
                if self.verbosity < len(self.loglevels) else logging.NOTSET)

    @classmethod
    def main(cls):
        """Execute command (as main entry point)"""
        cls().execute()

    def execute(self):
        """Execute command"""
        with self.device:
            getattr(self, self.args.subcommand)()

    def get(self):
        """Get variable(s)"""
        for name in self.args.variables:
            print(self.device.getvar(name))

    def set(self):
        """Set variable(s)"""
        for name, value in self.args.variables:
            self.device.setvar(name, value)

    def reset(self):
        """Reset device"""
        self.device.reset()

    def restore(self):
        """Restore configuration defaults"""
        self.device.restore_defaults(self.args.category)

    def ls(self):
        """List files"""
        print(self.device.list())

    def rm(self):
        """Remove file"""
        self.device.delete(self.args.filename)

    def mv(self):
        """Rename file"""
        self.device.rename(self.args.oldname, self.args.newname)

    def cat(self):
        """Get file contents"""
        print(self.device.download(self.args.filename).decode(), end='')

    def pull(self):
        """Download file"""
        outfile = self.args.output or self.args.filename
        name = os.path.basename(self.args.filename)
        with open(outfile, 'wb') as f:
            f.write(self.device.download(name))

    def push(self):
        """Upload file"""
        infile = self.args.input or self.args.filename
        name = os.path.basename(self.args.filename)
        with open(infile, 'rb') as f:
            self.device.upload(name, f.read())

    def upgrade(self):
        """Upgrade firmware"""
        firmware = ZebraFirmware(self.args.firmware)
        if not self.args.check:
            self.device.upgrade(firmware)

    def wifi(self):
        """Configure WiFi"""
        kwargs = {x: getattr(self.args, x)
                  for x in ('auth', 'country')
                  if getattr(self.args, x) is not None}
        self.device.wifi(self.args.essid, self.args.password, **kwargs)
        if self.args.reset:
            self.device.reset()



#!/usr/bin/env python3

"""Setup script"""

from setuptools import setup, find_packages

setup(
    name="zebconf",
    description="Zebra printer configuration",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author="Michael Brown",
    author_email="mbrown@fensystems.co.uk",
    url="https://github.com/unipartdigital/zebconf",
    license="GPLv2+",
    version="0.1.0",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: "
        "GNU General Public License v2 or later (GPLv2+)",
        "Programming Language :: Python :: 3",
        "Topic :: Printing",
    ],
    packages=find_packages(),
    install_requires=[
        'colorlog',
        'passlib',
        'pyusb',
    ],
    entry_points={
        'console_scripts': [
            'zebconf=zebconf.cli:ZebraCommand.main',
        ],
    },
)



"""Zebra printer device"""

import binascii
import logging
import re
from passlib.utils import pbkdf2
from .connection import (ZebraFileConnection, ZebraNetworkConnection,
                         ZebraUsbConnection)

logger = logging.getLogger(__name__)


class UnknownVariableError(LookupError):
    """Unknown configuration variable"""
    pass


class ZebraDevice(object):
    """A Zebra printer device"""

    MAX_RESPONSE_LEN = 16384
    """Maximum expected response length for any command"""

    DEFAULT_TIMEOUT = 2.0
    """Timeout used when reading from device"""

    def __init__(self, path=None, timeout=None):
        if path is None:
            self.conn = ZebraUsbConnection()
        elif '/' in path:
            self.conn = ZebraFileConnection(path)
        else:
            self.conn = ZebraNetworkConnection(path)
        self.timeout = timeout if timeout is not None else self.DEFAULT_TIMEOUT

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.conn.path)

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __getitem__(self, key):
        return self.getvar(key)

    def __setitem__(self, key, value):
        self.setvar(key, value)

    def open(self):
        """Open device"""
        self.conn.open()

    def close(self):
        """Close device"""
        self.conn.close()

    def write(self, data, printable=True):
        """Write data to device"""
        logger.debug('tx: %s',
                     data.decode().strip() if printable else data.hex())
        count = self.conn.write(data)
        assert count == len(data)

    def read(self, expect=None, printable=True):
        """Read data from device

        There is no consistent structure or delimiter for data read
        from the device.  The only viable approach is to read until an
        expected pattern match is seen, or a timeout occurs.
        """
        data = b''
        while True:
            frag = self.conn.read(self.MAX_RESPONSE_LEN, self.timeout)
            if not frag:
                break
            logger.debug('rx: %s', frag.decode() if printable else frag.hex())
            data += frag
            if expect is not None and re.fullmatch(expect, data):
                break
        if not data:
            raise TimeoutError
        return data

    def do(self, action, param=''):
        """Execute command"""
        self.write(b'! U1 do "%s" "%s"\r\n' %
                   (action.encode(), param.encode()))

    def setvar(self, name, value):
        """Set variable value"""
        self.write(b'! U1 setvar "%s" "%s"\r\n' %
                   (name.encode(), value.encode()))

    def setint(self, name, value):
        """Set integer variable value"""
        self.setvar(name, str(value))

    def setbool(self, name, value):
        """Set boolean variable value"""
        self.setvar(name, 'on' if value else 'off')

    def getvar(self, name):
        """Get variable value"""
        self.write(b'! U1 getvar "%s"\r\n' % name.encode())
        value = self.read(rb'".+?"').decode()
        if value == '"?"':
            raise UnknownVariableError(name)
        return value.strip('"')

    def getint(self, name):
        """Get integer variable value"""
        return int(self.getvar(name))

    def getbool(self, name):
        """Get boolean variable value"""
        value = self.getvar(name)
        if value == 'on':
            return True
        elif value == 'off':
            return False
        raise ValueError("%s: invalid value '%s'" % (name, value))

    def reset(self):
        """Reset device"""
        self.do('device.reset')

    def restore_defaults(self, category):
        """Restore configuration defaults"""
        self.do('device.restore_defaults', category)

    def list(self):
        """List files"""
        self.do('file.dir')
        return self.read().decode().strip('"')

    def delete(self, filename):
        """Remove file"""
        self.do('file.delete', filename)

    def rename(self, oldname, newname):
        """Rename file"""
        self.do('file.rename', '%s %s' % (oldname, newname))

    def download(self, filename):
        """Download file"""
        self.do('file.type', filename)
        return self.read(printable=False)

    def upload(self, filename, content):
        """Upload file"""
        self.write(b'! CISDFCRC16\r\n0000\r\n%s\r\n%08x\r\n0000\r\n%s' %
                   (filename.encode(), len(content), content), printable=False)

    def upgrade(self, firmware):
        """Upgrade firmware"""
        self.write(bytes(firmware))

    def wifi(self, essid, password, auth='wpa_psk', country=None):
        """Configure WiFi ESSID and password"""
        self.restore_defaults('wlan')
        self.setvar('wlan.essid', essid)
        if country is not None:
            self.setvar('wlan.country_code', country)
        getattr(self, 'wifi_%s' % auth)(essid, password)

    def wifi_wpa_psk(self, essid, password):
        """Configure WiFi ESSID and WPA-PSK password"""
        raw = pbkdf2.pbkdf2(password.encode(), essid.encode(), 4096, 32)
        psk = binascii.b2a_hex(raw).decode().upper()
        self.setbool('wlan.wpa.enable', True)
        self.setvar('wlan.wpa.authentication', 'psk')
        self.setvar('wlan.wpa.psk', psk)
